"""
Loads the opavote schema
"""

from rcvformats.schemas.base import GenericJsonSchema


#  pylint:disable=invalid-name
class SchemaV1_0(GenericJsonSchema):
    """ Schema for the initial version of the Universal RCV Tabulator """

    @property
    def schema_filename(self):
        return 'opavote.schema.json'

    def version(self):
        return "1.0"



"""
Interface and exceptions for all converters
"""

import abc

from rcvformats.common import utils
from rcvformats.schemas import universaltabulator


class CouldNotConvertException(Exception):
    """
    Raised when an unexpected error prevented conversion.
    This could be anything from an invalid input file, to unsupported data,
    to a bug in the software.
    """


class CouldNotOpenFileException(Exception):
    """
    Raised if the file could not be opened
    """


class Converter(abc.ABC):
    """ Interface for converters """

    def __init__(self):
        """ Initializes common data """
        self.ut_schema = universaltabulator.SchemaV0()

    def convert_to_ut_and_validate(self, filename_or_fileobj):
        """
        Calls :func:`~convert_to_ut`, then validates it with the Universal Tabulator schema.

        :param filename: A File object or filename
        :return: Guaranteed-valid Universal Tabulator data
        :raises CouldNotConvertException: If the conversion could not complete
        """
        try:
            ut_format = self.convert_to_ut(filename_or_fileobj)
            if not self.ut_schema.validate_data(ut_format):
                raise CouldNotConvertException(self.ut_schema.last_error())
        except Exception as unknown_error:
            raise CouldNotConvertException from unknown_error

        return ut_format

    def convert_to_ut(self, filename_or_fileobj):
        """
        Parses the file and returns the parsed data

        :param filename: A File object or filename
        :return: The Universal Tabulator representation of this data.\
                 Call :func:`~convert_to_ut_and_validate` to guarantee that \
                 it matches the Universal Tabulator schema.
        """
        if utils.is_file_obj(filename_or_fileobj):
            return self._convert_file_object_to_ut(filename_or_fileobj)
        if utils.is_filename(filename_or_fileobj):
            with open(filename_or_fileobj, 'r') as file_object:
                return self._convert_file_object_to_ut(file_object)
        raise CouldNotOpenFileException(f"Could not open {filename_or_fileobj}")

    @abc.abstractmethod
    def _convert_file_object_to_ut(self, file_object):
        """
        Just like func:`~convert_to_ut`, but only accepting a file object
        """


class GenericGuessAtTransferConverter(Converter):
    """
    If there are multiple candidates transferring their votes, we cannot know
    which candidates contributed to which transfers.
    Our best-effort guess is that they distributed their votes equally to all candidates.

    This base class lets you fill out partial data, leaving out tallyResults,
    and it will guess at the tallyResults for you.

    Note that it always knows the correct tally results if only one candidate is
    eliminated or elected in each round - the guessing only happens when multiple
    candidates are transferring their votes.
    """

    @classmethod
    def _weights_for_each_transfer(cls, eliminated_names, elected_names, vote_delta):
        """
        :param eliminated_names: the names of each eliminated candidate
        :param elected_names: the names of each elected candidate
        :param vote_delta: a dict mapping candidate name to vote difference
                           between this round and the next round
        :return: a dict mapping a transferring candidate's name to the weight they contributed
                to the overall transfer. In most cases, there will only be one candidate in
                transferring_candidates and the weight will be a simple 1.0.
        """
        # Names of both eliminated and elected candidates (elected may or may not have surpluses)
        transferring_candidates = eliminated_names + elected_names

        # How many votes are being transferred?
        votes_subtracted = -sum(d for d in vote_delta.values() if d < 0)

        weights = {
            name: -vote_delta[name] / votes_subtracted if votes_subtracted != 0 else 1
            for name in transferring_candidates
        }

        # Sanity check 1: weights must add up to 1
        if votes_subtracted != 0:
            if abs(sum(weights.values()) - 1.0) > 1e-8:
                votes_gained = sum(vote_delta.values())
                message = "Weights do not line up. This data is invalid. "\
                          f"In the round where {eliminated_names} were eliminated and "\
                          f"{elected_names} were elected, there were {votes_subtracted} "\
                          f"votes lost from the outgoing candidates, but {votes_gained} "\
                          "votes gained from continuing candidates. The number of votes must "\
                          "stay the same or decrease between rounds."
                raise ValueError(message)

        # Sanity check 2: can't gain more votes than you lost.
        # Exhausted ballots account for the inequality: lost ballots that were not gained anywhere
        votes_added = sum(d for d in vote_delta.values() if d > 0)
        assert votes_subtracted >= votes_added

        return weights

    @classmethod
    def _compute_vote_deltas_from_tally(cls, tally_this_round, tally_next_round):
        """
        Returns the vote deltas between this round and next round. The two tally arguments
        must match the tally format in the Universal Tabulator format.

        :param tally_this_round: A dict mapping candidate names to number of votes this round
        :param tally_next_round: A dict mapping candidate names to number of votes next round,\
                                 ensuring that any eliminated candidates are NOT present next\
                                 round, NOT that they just have zero votes.
        :return: A dict mapping candidate names in this round to how their votes changed
                 between this round and the next. Includes positive numbers (gained votes) and
                 negative numbers (lost votes via elimination or surplus transfer)
        """
        # Check who is no longer around next round
        eliminated_names = [name for name in tally_this_round if name not in tally_next_round]

        # Same as tally_next_round but includes 0 for eliminated candidates
        numvotes_next_round = {
            name: tally_next_round[name] if name not in eliminated_names else 0
            for name in tally_this_round
        }

        # Compute the difference for each candidate in this round
        return {
            to_name: numvotes_next_round[to_name] - tally_this_round[to_name]
            for to_name in tally_this_round
        }

    @classmethod
    def compute_vote_deltas_for_round(cls, ut_rounds_tally_only, round_i):
        """
        Gathers some data and passes it on to :func:`~_compute_vote_deltas_from_tally`

        :param ut_rounds_tally_only: Incomplete Universal Tabulator 'results' structure, \
                                     containing only 'tally' but not 'tallyResults'. \
                                     The tallies must be numbers, not strings.
        :param round_i: Computes delta between this round and the next
        :return: Value from :func:`~_compute_vote_deltas_from_tally`
        """
        if round_i == len(ut_rounds_tally_only) - 1:
            # Last round - no deltas
            return {}
        tally_this_round = ut_rounds_tally_only[round_i]['tally']
        tally_next_round = ut_rounds_tally_only[round_i + 1]['tally']
        return cls._compute_vote_deltas_from_tally(tally_this_round, tally_next_round)

    @classmethod
    def guess_at_tally_results(
            cls,
            eliminated_names,
            elected_names,
            vote_delta,
            allow_guessing=True):
        """
        Computes the tallyResult, the difference between this round and the next
        See the description of @_compute_tally_results to understand why, in the case of
        multiple winners, the best we can do is guess at the transfers here.

        :param eliminated_names: the names of each eliminated candidate
        :param elected_names: the names of each elected candidate
        :param vote_delta: a dict mapping candidate name to vote difference \
                           between this round and the next round.
        :param allow_guessing: Allow guessing of transfer data during batch elimination
        :return: The contents of the tallyResults dict
        """
        weights = cls._weights_for_each_transfer(eliminated_names, elected_names, vote_delta)

        # Loop over each eliminated or elected candidate and generate the transfers to
        # continuing candidates
        tally_results = []
        transfer_methods = {
            'eliminated': eliminated_names,
            'elected': elected_names
        }
        names_to_transfer_to = [name for name in vote_delta.keys() if
                                name not in eliminated_names and
                                name not in elected_names]
        for method, names in transfer_methods.items():
            for from_name in names:
                tally_result = {}
                tally_result[method] = from_name
                is_batch_elimination_round = len(names) > 1
                if not is_batch_elimination_round or allow_guessing:
                    transfers = {
                        to_name: vote_delta[to_name] * weights[from_name]
                        for to_name in names_to_transfer_to
                        if vote_delta[to_name] != 0
                    }
                else:
                    transfers = {}
                if transfers or method == 'eliminated':
                    # only add transfers on winners if they have actually been transferred
                    tally_result['transfers'] = transfers

                tally_results.append(tally_result)
        return tally_results



"""
Integration tests for conversions between file formats
"""

import os
import json
import nose

from rcvformats.conversions import automatic
from rcvformats.conversions import electionbuddy
from rcvformats.conversions import opavote
from rcvformats.conversions.ut_without_transfers import UTWithoutTransfersConverter


def _assert_conversion_correct(file_in, file_out, converter):
    """ Asserts that converter.convert_to_ut(file_in) = file_out """
    nose.tools.assert_dict_equal.__self__.maxDiff = None
    actual_data = converter.convert_to_ut_and_validate(file_in)
    with open(file_out, 'r') as file_obj:
        expected_data = json.load(file_obj)
    nose.tools.assert_dict_equal(actual_data, expected_data)


def _files_in_dir(input_dir):
    """ Yields all CSV and json files in dir """
    for filename in os.listdir(input_dir):
        if not filename.endswith('.json') and not filename.endswith('.csv'):
            continue
        yield filename


def _assert_auto_gives_same_result_as(input_dir, direct_converter):
    """
    Asserts that the AutomaticConverter creates the same results as the provided one
    for every file in the provided directory
    """
    auto_converter = automatic.AutomaticConverter()

    for filename in _files_in_dir(input_dir):
        filepath = os.path.join(input_dir, filename)
        auto_data = auto_converter.convert_to_ut_and_validate(filepath)
        expected_data = direct_converter.convert_to_ut_and_validate(filepath)

        nose.tools.assert_dict_equal(auto_data, expected_data)


def test_electionbuddy_conversions_succeed():
    """ Converts electionbuddy CSV to standard format """
    filenames = [
        'testdata/inputs/electionbuddy/standard.csv',
        'testdata/inputs/electionbuddy/multiwinner.csv',
        'testdata/inputs/electionbuddy/without-abstentions.csv'
    ]
    converter = electionbuddy.ElectionBuddyConverter()
    for filename in filenames:
        converter.convert_to_ut_and_validate(filename)


def test_opavote_conversion_accurate():
    """ Converts opavote JSON to standard format """
    file_in = 'testdata/inputs/opavote/fairvote.json'
    file_out = 'testdata/conversions/from-opavote.json'
    converter = opavote.OpavoteConverter()
    _assert_conversion_correct(file_in, file_out, converter)


def test_electionbuddy_conversion_accurate():
    """ Converts electionbuddy CSV to the standard format """
    file_in = 'testdata/inputs/electionbuddy/standard.csv'
    file_out = 'testdata/conversions/from-electionbuddy.json'
    converter = electionbuddy.ElectionBuddyConverter()
    _assert_conversion_correct(file_in, file_out, converter)

    file_in = 'testdata/inputs/electionbuddy/standard-with-threshold.csv'
    file_out = 'testdata/conversions/from-electionbuddy-with-threshold.json'
    converter = electionbuddy.ElectionBuddyConverter()
    _assert_conversion_correct(file_in, file_out, converter)


def test_automatic_conversions_universal_tabulator():
    """ Tests that the automatic conversion works when given Universal Tabulator data """
    converter = automatic.AutomaticConverter()
    input_dir = 'testdata/inputs/universal-tabulator'

    for filename in _files_in_dir(input_dir):
        filepath = os.path.join(input_dir, filename)
        output_data = converter.convert_to_ut_and_validate(filepath)

        # Must be unchanged
        with open(filepath, 'r') as input_file:
            input_data = json.load(input_file)

        nose.tools.assert_dict_equal(input_data, output_data)


def test_automatic_conversions_opavote():
    """ Tests that the automatic conversion works when given Opavote data """
    converter = opavote.OpavoteConverter()
    input_dir = 'testdata/inputs/opavote'
    _assert_auto_gives_same_result_as(input_dir, converter)


def test_automatic_conversions_electionbuddy():
    """ Tests that the automatic conversion works when given ElectionBuddy data """
    converter = electionbuddy.ElectionBuddyConverter()
    input_dir = 'testdata/inputs/electionbuddy'
    _assert_auto_gives_same_result_as(input_dir, converter)


def _does_all_single_elim_have_transfer_data(data):
    for result in data['results']:
        tally_results = result['tallyResults']
        if len(tally_results) != 1:
            continue

        tally_result = tally_results[0]
        if 'elected' in tally_result:
            continue

        if not tally_result['transfers']:
            return False
    return True


def _does_all_batch_elim_have_transfer_data(data):
    for result in data['results']:
        tally_results = result['tallyResults']
        if len(tally_results) == 1:
            continue

        for tally_result in tally_results:
            if 'elected' in tally_result:
                continue
            if not tally_result['transfers']:
                return False
    return True


def test_add_xfer_without_fake_data():
    """ Tests allow_guessing=False in UT Without Transfers """
    converter = UTWithoutTransfersConverter(allow_guessing=False)
    input_filename = 'testdata/inputs/ut-without-transfers/nyc-batch-elim.json'
    with_transfers = converter.convert_to_ut(input_filename)
    assert _does_all_single_elim_have_transfer_data(with_transfers)
    assert not _does_all_batch_elim_have_transfer_data(with_transfers)


def test_add_xfer_with_fake_data():
    """ Tests allow_guessing=True in UT Without Transfers """
    converter = UTWithoutTransfersConverter(allow_guessing=True)
    input_filename = 'testdata/inputs/ut-without-transfers/nyc-batch-elim.json'
    with_transfers = converter.convert_to_ut(input_filename)
    assert _does_all_single_elim_have_transfer_data(with_transfers)
    assert _does_all_batch_elim_have_transfer_data(with_transfers)



"""
Loads supported schemas (currently only one: the Universal Tabulator schema)
"""

import abc
import json
import os

import jsonschema

from rcvformats.common import utils


class Schema(abc.ABC):
    """
    A single version of a single schema
    """

    def __init__(self):
        self._last_error = None

    @abc.abstractmethod
    def version(self):
        """
        The version number of this schema

        :return: A string represting the version number
        """

    def validate(self, filename_or_fileobj):
        """
        Validates that the file matches the expected schema

        :param filename_or_fileobj: The JSON filename or file object for the tabulated results
        :return: whether or not the validation failed
        """
        if utils.is_file_obj(filename_or_fileobj):
            return self._validate_file_object(filename_or_fileobj)
        if utils.is_filename(filename_or_fileobj):
            with open(filename_or_fileobj, 'r') as file_object:
                return self._validate_file_object(file_object)
        # Couldn't open the file at all
        self._last_error = TypeError("Couldn't open file")
        return False

    @abc.abstractmethod
    def _validate_file_object(self, file_object):
        """
        Implements the bulk of func:`~validate`
        """

    def last_error(self):
        """
        If validate() failed, this method will provide more detailed information
        on the error. The details vary by class type, though it will always be of type
        Exception

        :return: Exception with additional information on why the validation failed
        """
        assert self._last_error is None or isinstance(self._last_error, Exception)
        return self._last_error


class GenericJsonSchema(Schema):
    """ Base class for a JSON Schema """
    @property
    @abc.abstractmethod
    def schema_filename(self):
        """ The JSON Schema filename """

    def __init__(self):
        filename = self.schema_filename
        filepath = os.path.join(self._get_jsonschema_directory(), filename)
        with open(filepath, 'r') as file_object:
            self.schema = json.load(file_object)

        super().__init__()

    @classmethod
    def _get_jsonschema_directory(cls):
        return os.path.join(os.path.dirname(__file__), '..', 'jsonschemas')

    def _validate_file_object(self, file_object):
        """ Opens the file and runs :func:`~validate_data` """
        try:
            data = json.load(file_object)
        except json.decoder.JSONDecodeError as error:
            self._last_error = error
            return False

        return self.validate_data(data)

    def validate_data(self, data):
        """
        Validates that the data matches the schema. If invalid, more data may be available
        by calling last_error()

        :param data: The input dictionary
        :return: Whether or not the data matches the schema
        """
        try:
            jsonschema.validate(data, self.schema)
            return True
        except jsonschema.exceptions.ValidationError as error:
            self._last_error = error
            return False



"""
Attempts to convert any file to the standard format.
If you know the file format, you should not use this - it will
loop through all schemas which will be needlessly slow.
"""

import json

from rcvformats.conversions.base import CouldNotConvertException
from rcvformats.conversions.electionbuddy import ElectionBuddyConverter
from rcvformats.conversions.opavote import OpavoteConverter
from rcvformats.schemas.electionbuddy import SchemaV0 as ElectionBuddySchema
from rcvformats.schemas.opavote import SchemaV1_0 as OpavoteSchema
from rcvformats.conversions.base import Converter


class AutomaticConverter(Converter):
    """ Interface for converters """

    def __init__(self):
        # In order of likelihood of a hit - just my guess
        self.additional_schemas = [
            (ElectionBuddySchema(), ElectionBuddyConverter),
            (OpavoteSchema(), OpavoteConverter)
        ]

        super().__init__()

    def _convert_file_object_to_ut(self, file_object):
        # If it matches the schema already, load and return the data
        if self.ut_schema.validate(file_object):
            file_object.seek(0)
            return json.load(file_object)

        # Otherwise, loop through each schema, and if it matches, run the converter
        for schema, converter in self.additional_schemas:
            file_object.seek(0)
            matches_schema = schema.validate(file_object)
            if matches_schema:
                file_object.seek(0)
                return converter().convert_to_ut(file_object)

        # If it failed, accumulate all errors from schemas
        err = CouldNotConvertException("Could not find a compatible parser")
        err.args += (self.ut_schema.last_error(),)
        for schema, _ in self.additional_schemas:
            err.args += (schema.last_error(),)
        raise err



"""
Loads the universal tabulator schema
"""

from rcvformats.schemas.base import GenericJsonSchema


class SchemaV0(GenericJsonSchema):
    """ Schema for the initial version of the Universal RCV Tabulator """

    @property
    def schema_filename(self):
        return 'universaltabulator.schema.json'

    def version(self):
        return "Unversioned:V0"



"""
Add transfers to an otherwise-valid Universal Tabulator format
"""

import json

from rcvformats.conversions.base import GenericGuessAtTransferConverter


class UTWithoutTransfersConverter(GenericGuessAtTransferConverter):
    """
    Reads an UT-formatted JSON file that is missing "transfers"
    """

    def __init__(self, allow_guessing=True):
        """
        @param allow_guessing During batch elimination, should we guess at
                              where the votes went (distributing them proportionally)
                              or leave this blank?
        """
        self.allow_guessing = allow_guessing
        super().__init__()

    def _convert_file_object_to_ut(self, file_object):
        data = json.load(file_object)

        self._convert_tally_string_to_decimal(data['results'])
        self._fill_in_tallyresults(data['results'])
        return data

    @classmethod
    def _get_eliminated_names(cls, rounds, round_i):
        """ Reads each tallyResult, returns any eliminated name """
        results = rounds[round_i]['tallyResults']
        return [d['eliminated'] for d in results if 'eliminated' in d]

    @classmethod
    def _get_elected_names(cls, rounds, round_i):
        """ Reads each tallyResult, returns any elected name """
        results = rounds[round_i]['tallyResults']
        return [d['elected'] for d in results if 'elected' in d]

    @classmethod
    def _convert_tally_string_to_decimal(cls, rounds):
        for round_data in rounds:
            for person in round_data['tally']:
                round_data['tally'][person] = float(round_data['tally'][person])

    def _fill_in_tallyresults(self, rounds):
        """ Fill out rounds['tallyResults'] based on rounds['tally'] """
        for round_i, _ in enumerate(rounds):
            # Get who was elected and eliminated
            eliminated_names = self._get_eliminated_names(rounds, round_i)
            elected_names = self._get_elected_names(rounds, round_i)

            # Get how the votes change between this round and next
            vote_delta = self.compute_vote_deltas_for_round(rounds, round_i)

            # Use that to compute the tallyResults structure
            tally_results = self.guess_at_tally_results(
                eliminated_names, elected_names, vote_delta, self.allow_guessing)

            # Store it, completing the structure for this round
            rounds[round_i]['tallyResults'] = tally_results



"""
Tests that our example JSONs all pass validation
"""

import json
import tempfile
import os

from rcvformats.schemas import electionbuddy
from rcvformats.schemas import universaltabulator
from rcvformats.schemas import opavote


def _validate_files_in_dir_with(directory, schema):
    for filename in os.listdir(directory):
        if not schema.validate(os.path.join(directory, filename)):
            raise schema.last_error()


def _modify_json_with(json_filename, modifier_func):
    """
    Given a modifier_func which modifies the data in the given file,
    updates the json file and returns a Tempfile holding the new data
    """
    # Update the sidecar data
    with open(json_filename, 'r+') as file_obj:
        data = json.load(file_obj)
        modifier_func(data)

    # Write it to a tempfile
    modified_json_tempfile = tempfile.NamedTemporaryFile()  # pylint: disable=consider-using-with
    with open(modified_json_tempfile.name, 'w') as file_obj:
        json.dump(data, file_obj)
    return modified_json_tempfile


def test_universal_tabulator_formats_valid():
    """ Verifies that the example formats are valid """
    schema = universaltabulator.SchemaV0()
    _validate_files_in_dir_with('testdata/inputs/universal-tabulator', schema)


def test_opavote_formats_valid():
    """ Verifies that the example formats are valid """
    schema = opavote.SchemaV1_0()
    _validate_files_in_dir_with('testdata/inputs/opavote', schema)


def test_electionbuddy_formats_valid():
    """ Verifies that the example formats are valid """
    schema = electionbuddy.SchemaV0()
    _validate_files_in_dir_with('testdata/inputs/electionbuddy', schema)


def test_empty_threshold_not_valid():
    """ (Regression) Verifies that empty string is not valid for threshold """
    badjson_tempfile = _modify_json_with('testdata/inputs/universal-tabulator/one-round.json',
                                         lambda d: d['config'].update({'threshold': ''}))
    schema = universaltabulator.SchemaV0()
    assert not schema.validate(badjson_tempfile.name)


def test_threshold_is_optional():
    """ Verifies that threshold may be nonexistent """
    badjson_tempfile = _modify_json_with('testdata/inputs/universal-tabulator/one-round.json',
                                         lambda d: d['config'].pop('threshold'))
    schema = universaltabulator.SchemaV0()
    assert schema.validate(badjson_tempfile.name)



"""
A collection of shared helper utilities
"""

import io
import os


def is_file_obj(filename_or_fileobj):
    """ Is the given argument a File-like object? """
    if isinstance(filename_or_fileobj, io.IOBase):
        return True
    if hasattr(filename_or_fileobj, 'read') and hasattr(filename_or_fileobj, 'seek'):
        return True
    return False


def is_filename(filename_or_fileobj):
    """ Is the given argument a filename? """
    return os.path.isfile(filename_or_fileobj)



"""
Creates the rcvformats package
"""

import os
from setuptools import setup, find_packages

version = os.environ.get('TAG_FROM_GITHUB_ACTIONS', '0.0.0')

def readme():
    """ Returns contents of README.md """
    with open("README.md", 'r') as file_obj:
        return file_obj.read()

setup(
    name='rcvformats',
    version=version,
    description='Schema validations, migrations, and conversions to '\
                'standardize the Ranked Choice Voting ecosystem',
    long_description=readme(),
    long_description_content_type="text/markdown",
    url='https://github.com/artoonie/rcvformats',
    author="Armin Samii",
    author_email="armin.samii@gmail.com",
    packages=find_packages(),
    include_package_data=True,
    python_requires='>=3',
    install_requires=[
        'jsonschema'
    ],
    entry_points={
        'console_scripts': ['rcvformats=rcvformats.bin.cli:main'],
    }
)



"""
Reads an ElectionBuddy CSV results file, writes to the standard format
"""

import json

from rcvformats.conversions.base import GenericGuessAtTransferConverter


class OpavoteConverter(GenericGuessAtTransferConverter):
    """
    Reads an opavote-formatted JSON file.
    """

    @classmethod
    def _get_eliminated_names(cls, rounds, candidate_names, round_i):
        """
        Opavote format places losses on the round after they happen, whereas
        the Universal Tabulator format places it on the previous round.
        This function looks at the next round for its data. A corellary is that
        there are no eliminations on the first round, which is what I believe to
        always be the case anyway.

        :param rounds: rounds data, direct from the Opavote format
        :param candidate_names: in-order names
        :param round_i: the current round (we'll look at round_i+1)
        :return: list of names that were eliminated
        """
        if round_i == len(rounds) - 1:
            ids = rounds[-1]['losers']
        else:
            ids = rounds[round_i + 1]['losers']
        return [candidate_names[i] for i in ids]

    @classmethod
    def _get_elected_names(cls, rounds, candidate_names, round_i):
        """
        :param candidate_names: in-order names
        :param rounds: rounds data, direct from the Opavote format
        :param round_i: the current round
        :return: list of names that were elected
        """
        ids = rounds[round_i]['winners']
        return [candidate_names[i] for i in ids]

    @classmethod
    def _votes_on_round(cls, candidate_i, rounds, round_i):
        """ Number of votes the corresponding candidate had on round_i """
        return rounds[round_i]['count'][candidate_i]

    def _convert_file_object_to_ut(self, file_object):
        data = json.load(file_object)

        threshold = sum(data['rounds'][-1]['count']) / (data['n_seats'] + 1)
        ut_config = {
            'contest': data['title'],
            'threshold': threshold
        }

        # Fill out rounds['tally']
        rounds = data['rounds']
        candidate_names = data['candidates']
        ut_rounds = []
        for round_i in range(len(rounds)):
            ut_round = {
                'round': round_i + 1,
                'tally': {}
            }
            for candidate_i, _ in enumerate(candidate_names):
                votes = self._votes_on_round(candidate_i, rounds, round_i)
                name = candidate_names[candidate_i]
                ut_round['tally'][name] = votes
            ut_rounds.append(ut_round)

        self._fill_in_tallyresults(rounds, candidate_names, ut_rounds)
        self._remove_eliminated_candidates_from_tally(rounds, candidate_names, ut_rounds)

        return {'config': ut_config, 'results': ut_rounds}

    @classmethod
    def _remove_eliminated_candidates_from_tally(cls, rounds, candidate_names, ut_rounds):
        for round_i in range(1, len(rounds)):
            already_eliminated = cls._get_eliminated_names(rounds, candidate_names, round_i - 1)
            for candidate_name in already_eliminated:
                del ut_rounds[round_i]['tally'][candidate_name]

    @classmethod
    def _fill_in_tallyresults(cls, rounds, candidate_names, ut_rounds):
        """ Fill out rounds['tallyResults'] based on rounds['tally'] """
        already_eliminated = set()
        for round_i, _ in enumerate(rounds):
            # Get who was elected and eliminated
            eliminated_names = cls._get_eliminated_names(rounds, candidate_names, round_i)
            elected_names = cls._get_elected_names(rounds, candidate_names, round_i)
            # Opavote accumulates eliminated candidates. Clean that up.
            eliminated_names = [n for n in eliminated_names if n not in already_eliminated]
            already_eliminated.update(eliminated_names)
            # Get how the votes change between this round and next
            vote_delta = cls.compute_vote_deltas_for_round(ut_rounds, round_i)
            # Use that to compute the tallyResults structure
            tally_results = cls.guess_at_tally_results(eliminated_names, elected_names, vote_delta)
            # Store it, completing the structure for this round
            ut_rounds[round_i]['tallyResults'] = tally_results
